@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-# Copyright (C) 1998-2019 by the Free Software Foundation, Inc.
+# Copyright (C) 1998-2021 by the Free Software Foundation, Inc.
 #
 # This file is part of Postorius.
 #
@@ -37,7 +37,8 @@ from django.utils.translation import gettext as _
 from django.views.decorators.http import require_http_methods
 
 from allauth.account.models import EmailAddress
-from django_mailman3.lib.mailman import get_mailman_client
+from django_mailman3.lib.mailman import (
+    get_mailman_client, get_mailman_user, get_mailman_user_id)
 from django_mailman3.lib.paginator import MailmanPaginator, paginate
 from django_mailman3.models import MailDomain
 from django_mailman3.signals import (
@@ -48,19 +49,31 @@ from postorius.auth.decorators import (
 from postorius.auth.mixins import ListOwnerMixin
 from postorius.forms import (
     AlterMessagesForm, ArchiveSettingsForm, BounceProcessingForm,
-    DigestSettingsForm, DMARCMitigationsForm, ListAnonymousSubscribe,
-    ListAutomaticResponsesForm, ListHeaderMatchForm, ListHeaderMatchFormset,
-    ListIdentityForm, ListMassRemoval, ListMassSubscription, ListNew,
-    ListSubscribe, MemberForm, MemberModeration, MemberPolicyForm,
-    MessageAcceptanceForm, MultipleChoiceForm, UserPreferences)
-from postorius.forms.list_forms import ACTION_CHOICES
-from postorius.models import Domain, List, Mailman404Error, Style
+    ChangeSubscriptionForm, DigestSettingsForm, DMARCMitigationsForm,
+    ListAnonymousSubscribe, ListAutomaticResponsesForm, ListHeaderMatchForm,
+    ListHeaderMatchFormset, ListIdentityForm, ListMassRemoval,
+    ListMassSubscription, ListNew, ListSubscribe, MemberForm, MemberModeration,
+    MemberPolicyForm, MessageAcceptanceForm, MultipleChoiceForm,
+    UserPreferences)
+from postorius.forms.fields import (
+    DELIVERY_MODE_CHOICES, DELIVERY_STATUS_CHOICES)
+from postorius.forms.list_forms import ACTION_CHOICES, TokenConfirmForm
+from postorius.models import (
+    Domain, List, Mailman404Error, Style, SubscriptionMode)
+from postorius.utils import get_member_or_nonmember, set_preferred
 from postorius.views.generic import MailingListView, bans_view
 
 
 logger = logging.getLogger(__name__)
 
 
+#: DeliveryStatus field values that an admin cannot set.
+DISABLED_DELIVERY_STATUS_CHOICES_ADMIN = ['by_bounces']
+
+DELIVERY_MODE_DICT = dict(DELIVERY_MODE_CHOICES)
+DELIVERY_STATUS_DICT = dict(DELIVERY_STATUS_CHOICES)
+
+
 class TokenOwner:
     """Who 'owns' the token returned from the registrar?"""
     subscriber = 'subscriber'
@@ -114,7 +127,7 @@ class ListMembersViews(ListOwnerMixin, MailingListView):
             request.GET.get('count', 25),
             paginator_class=MailmanPaginator)
         context['page_subtitle'] = '({})'.format(
-            context['members'].paginator.count)
+            context['members'].object_list.total_size)
         context['form_action'] = _('Add {}'.format(role))
         if context['query']:
             context['empty_error'] = _(
@@ -177,7 +190,7 @@ class ListMembersViews(ListOwnerMixin, MailingListView):
         if role not in self.allowed_roles:
             return redirect('list_members', list_id, 'member')
 
-        if role in ('member'):
+        if role in ('member', ):
             return self._member_post(request, role)
         else:
             return self._non_member_post(request, role)
@@ -188,20 +201,29 @@ class ListMembersViews(ListOwnerMixin, MailingListView):
 def list_member_options(request, list_id, email):
     template_name = 'postorius/lists/memberoptions.html'
     mm_list = List.objects.get_or_404(fqdn_listname=list_id)
+    # If the role is specified explicitly, use that, otherwise the value is
+    # None and is equivalent to the value not being specified and the lookup
+    # happens using only email. This can cause issues when the email is
+    # subscribed as both a Non-Member and Owner/Moderator returning the wrong
+    # Member object.
+    role = request.GET.get('role')
     try:
-        mm_member = mm_list.find_members(address=email)[0]
+        mm_member = mm_list.find_members(address=email, role=role)[0]
         member_prefs = mm_member.preferences
     except (ValueError, IndexError):
         raise Http404(_('Member does not exist'))
     except Mailman404Error:
         return render(request, template_name, {'nolists': 'true'})
+
     initial_moderation = dict([
         (key, getattr(mm_member, key)) for key in MemberModeration.base_fields
         ])
+
     if request.method == 'POST':
         if request.POST.get("formname") == 'preferences':
             preferences_form = UserPreferences(
-                request.POST, preferences=member_prefs)
+                request.POST, preferences=member_prefs,
+                disabled_delivery_choices=DISABLED_DELIVERY_STATUS_CHOICES_ADMIN)  # noqa:E501
             if preferences_form.is_valid():
                 try:
                     preferences_form.save()
@@ -233,7 +255,9 @@ def list_member_options(request, list_id, email):
                                                 "settings have been updated."))
                 return redirect('list_member_options', list_id, email)
     else:
-        preferences_form = UserPreferences(preferences=member_prefs)
+        preferences_form = UserPreferences(
+            preferences=member_prefs,
+            disabled_delivery_choices=DISABLED_DELIVERY_STATUS_CHOICES_ADMIN)
         moderation_form = MemberModeration(initial=initial_moderation)
     return render(request, template_name, {
         'mm_member': mm_member,
@@ -251,6 +275,7 @@ class ListSummaryView(MailingListView):
         data = {'list': self.mailing_list,
                 'user_subscribed': False,
                 'subscribed_address': None,
+                'subscribed_preferred': False,
                 'public_archive': False,
                 'hyperkitty_enabled': False}
         if self.mailing_list.settings['archive_policy'] == 'public':
@@ -260,28 +285,50 @@ class ListSummaryView(MailingListView):
             # avoid systematic test failure when HyperKitty is installed
             # (missing VCR request, see below).
             list(self.mailing_list.archivers)
-        if ('hyperkitty' in settings.INSTALLED_APPS and           # noqa: W504
-                'hyperkitty' in self.mailing_list.archivers and   # noqa: W504
-                self.mailing_list.archivers['hyperkitty']):
+        archivers = self.mailing_list.archivers
+        if ('hyperkitty' in settings.INSTALLED_APPS and          # noqa: W504
+                'hyperkitty' in archivers and                    # noqa: W504
+                archivers['hyperkitty']):                       # noqa: W504
             data['hyperkitty_enabled'] = True
         if request.user.is_authenticated:
             user_emails = EmailAddress.objects.filter(
                 user=request.user, verified=True).order_by(
                 "email").values_list("email", flat=True)
             pending_requests = [r['email'] for r in self.mailing_list.requests]
+
+            subscriptions = []
             for address in user_emails:
                 if address in pending_requests:
                     data['user_request_pending'] = True
-                    break
+                    continue
                 try:
-                    self.mailing_list.get_member(address)
+                    member = self.mailing_list.get_member(address)
                 except ValueError:
                     pass
                 else:
-                    data['user_subscribed'] = True
-                    data['subscribed_address'] = address
-                    break  # no need to test more addresses
-            data['subscribe_form'] = ListSubscribe(user_emails)
+                    subscriptions.append({
+                        'subscribed_address': address,
+                        'subscriber': member.member_id,
+                        'subscribed_preferred': bool(
+                            member.subscription_mode ==
+                            SubscriptionMode.as_user.name),
+                        'delivery_mode': DELIVERY_MODE_DICT.get(
+                            member.preferences.get('delivery_mode')),
+                        'delivery_status': DELIVERY_STATUS_DICT.get(
+                            member.preferences.get('delivery_status'))
+                    })
+            data['user_subscriptions'] = subscriptions
+            data['user_subscribed'] = bool(subscriptions)
+
+            mm_user = get_mailman_user(request.user)
+            primary_email = None
+            if mm_user.preferred_address is None:
+                primary_email = set_preferred(request.user, mm_user)
+            else:
+                primary_email = mm_user.preferred_address.email
+            data['subscribe_form'] = ListSubscribe(
+                user_emails, user_id=mm_user.user_id,
+                primary_email=primary_email)
         else:
             user_emails = None
             data['anonymous_subscription_form'] = ListAnonymousSubscribe()
@@ -289,8 +336,81 @@ class ListSummaryView(MailingListView):
 
 
 class ChangeSubscriptionView(MailingListView):
-    """Change mailing list subscription
-    """
+    """Change mailing list subscription"""
+
+    @staticmethod
+    def _is_email(subscriber):
+        """Is the subscriber an email address or uuid."""
+        return '@' in subscriber
+
+    def _change_subscription(self, member, subscriber):
+        """Switch subscriptions to a new email/user.
+
+        :param member: The currently subscriber Member.
+        :param subscriber: New email or user_id to switch subscription to.
+        """
+        # If the Membership is via preferred address and we are switching to an
+        # email or vice versa, we have to do the subscribe-unsubscribe dance.
+        # If the subscription is via an address, then we can simply PATCH the
+        # address in the Member resource.
+        if (member.subscription_mode == SubscriptionMode.as_address.name and
+                self._is_email(subscriber)):
+            member.address = subscriber
+            member.save()
+            return
+        # Keep a copy of the preferences so we can restore them after the
+        # Member switches addresses.
+        if member.preferences:
+            prefs = member.preferences
+        member.unsubscribe()
+        # Unless the subscriber is an an email and it is not-verified, we
+        # should get a Member resource here as response. We should never get to
+        # here with subscriber being an un-verified email.
+        new_member = self.mailing_list.subscribe(
+            subscriber,
+            # Since the email is already verified in Postorius.
+            pre_confirmed=True,
+            # Since this user was already a Member, simply switching Email
+            # addresses shouldn't require another approval.
+            pre_approved=True)
+        self._copy_preferences(new_member.preferences, prefs)
+
+    def _copy_preferences(self, member_pref, old_member_pref):
+        """Copy the preferences of the old member to new one.
+
+        :param member_pref: The new member's preference to copy preference to.
+        :param old_member_pref: The old_member's preferences to copy
+            preferences from.
+        """
+        # We can't simply switch the Preferences object, so we copy values from
+        # previous one to the new one.
+        for prop in old_member_pref._properties:
+            val = old_member_pref.get(prop)
+            if val:
+                member_pref[prop] = val
+        member_pref.save()
+
+    def _get_membership(self, mm_user, member_id):
+        """Get current memberships of the Mailman user."""
+        for sub in mm_user.subscriptions:
+            if sub.member_id == member_id:
+                return sub
+        return None
+
+    def _is_subscribed(self, member, subscriber, mm_user):
+        """Check if the member is same as the subscriber.
+
+        If they are not:
+        - If subscriber is 'user_id' check if member's user_id is same as
+           subscriber.
+        - if subscriber is address, check if member's address is same as
+           subscriber.
+        """
+        return (
+            (member.subscription_mode == SubscriptionMode.as_address.name and
+             member.email == subscriber) or
+            (member.subscription_mode == SubscriptionMode.as_user.name and
+             member.user.user_id == subscriber))
 
     @method_decorator(login_required)
     def post(self, request, list_id):
@@ -298,41 +418,50 @@ class ChangeSubscriptionView(MailingListView):
             user_emails = EmailAddress.objects.filter(
                 user=request.user, verified=True).order_by(
                 "email").values_list("email", flat=True)
-            form = ListSubscribe(user_emails, request.POST)
-            # Find the currently subscribed email
-            old_email = None
-            for address in user_emails:
-                try:
-                    self.mailing_list.get_member(address)
-                except ValueError:
-                    pass
-                else:
-                    old_email = address
-                    break  # no need to test more addresses
-            assert old_email is not None
+            mm_user = get_mailman_user(request.user)
+            primary_email = None
+            if mm_user.preferred_address is not None:
+                primary_email = mm_user.preferred_address.email
+            form = ChangeSubscriptionForm(
+                user_emails, mm_user.user_id, primary_email, request.POST)
+
             if form.is_valid():
-                email = form.cleaned_data['email']
-                if old_email == email:
+                # Step 1: Get the source membership user wants to switch from.
+                member = self._get_membership(
+                    mm_user, form.cleaned_data['member_id'])
+                if member is None:
+                    # Source membership doesn't exist, use should use subscribe
+                    # form instead.
+                    messages.error(_('You are not subscribed to {}.').format(
+                        self.mailing_list.fqdn_listname))
+                    # Not sure what the right place to return to here, since
+                    # there is no on_get resource here. Return to
+                    # per-subscription-preferences where the user can choose
+                    # subscription to change.
+                    return redirect('user_subscription_preferences')
+                subscriber = form.cleaned_data['subscriber']
+                # Step 2: Check if the source and destination subscriber are
+                # the same. This means, if the subscriber is a user_id, check
+                # they match and if the subscriber is a email address, check if
+                # they match.
+                if self._is_subscribed(member, subscriber, mm_user):
                     messages.error(request, _('You are already subscribed'))
                 else:
-                    self.mailing_list.unsubscribe(old_email)
-                    # Since the action is done via the web UI, no email
-                    # confirmation is needed.
-                    response = self.mailing_list.subscribe(
-                        email, pre_confirmed=True)
-                    if (type(response) == dict and                # noqa: W504
-                            response.get('token_owner') == TokenOwner.moderator):  # noqa: E501
-                        messages.success(
-                            request, _('Your request to change the email for'
-                                       ' this subscription was submitted and'
-                                       ' is waiting for moderator approval.'))
-                    else:
-                        messages.success(request,
-                                         _('Subscription changed to %s') %
-                                         email)
+                    self._change_subscription(member, subscriber)
+                    # If the subscriber is user_id (i.e. not an email
+                    # address), Show 'Primary Address ()' in the success
+                    # message instead of weird user id.
+                    if not self._is_email(subscriber):
+                        subscriber = _(
+                            'Primary Address ({})').format(primary_email)
+                    messages.success(
+                        request,
+                        _('Subscription changed to {}').format(subscriber))
             else:
-                messages.error(request,
-                               _('Something went wrong. Please try again.'))
+                messages.error(
+                    request,
+                    _('Something went wrong. Please try again. {} {}').format(
+                        form.errors, form.fields['subscriber'].choices))
         except HTTPError as e:
             messages.error(request, e.msg)
         return redirect('list_summary', self.mailing_list.list_id)
@@ -353,11 +482,22 @@ class ListSubscribeView(MailingListView):
             user_emails = EmailAddress.objects.filter(
                 user=request.user, verified=True).order_by(
                 "email").values_list("email", flat=True)
-            form = ListSubscribe(user_emails, request.POST)
+            mm_user = get_mailman_user(request.user)
+            primary_email = None
+            if mm_user.preferred_address is not None:
+                primary_email = mm_user.preferred_address.email
+            form = ListSubscribe(
+                user_emails, mm_user.user_id, primary_email, request.POST)
             if form.is_valid():
-                email = request.POST.get('email')
+                subscriber = form.cleaned_data.get('subscriber')
+                display_name = form.cleaned_data.get('display_name')
+                delivery_mode = form.cleaned_data.get('delivery_mode')
+                delivery_status = form.cleaned_data.get('delivery_status')
                 response = self.mailing_list.subscribe(
-                    email, pre_verified=True, pre_confirmed=True)
+                    subscriber, display_name,
+                    pre_verified=True, pre_confirmed=True,
+                    delivery_mode=delivery_mode,
+                    delivery_status=delivery_status)
                 if (type(response) == dict and                   # noqa: W504
                         response.get('token_owner') == TokenOwner.moderator):
                     messages.success(
@@ -392,8 +532,10 @@ class ListAnonymousSubscribeView(MailingListView):
             form = ListAnonymousSubscribe(request.POST)
             if form.is_valid():
                 email = form.cleaned_data.get('email')
-                self.mailing_list.subscribe(email, pre_verified=False,
-                                            pre_confirmed=False)
+                display_name = form.cleaned_data.get('display_name')
+                self.mailing_list.subscribe(
+                    email, display_name,
+                    pre_verified=False, pre_confirmed=False)
                 messages.success(request, _('Please check your inbox for '
                                             'further instructions'))
             else:
@@ -411,10 +553,31 @@ class ListUnsubscribeView(MailingListView):
     @method_decorator(login_required)
     def post(self, request, *args, **kwargs):
         email = request.POST['email']
+        # Verify the user actually controls this email, should
+        # return 1 if the user owns the email, 0 otherwise.
+        found_email = EmailAddress.objects.filter(
+            user=request.user, email=email, verified=True).count()
+        if found_email == 0:
+            messages.error(
+                request,
+                _('You can only unsubscribe yourself.'))
+            return redirect('list_summary', self.mailing_list.list_id)
+        if self._has_pending_unsub_req(email):
+            messages.error(
+                request,
+                _('You have a pending unsubscription request waiting for'
+                  ' moderator approval.'))
+            return redirect('list_summary', self.mailing_list.list_id)
         try:
-            self.mailing_list.unsubscribe(email)
-            messages.success(request, _('%s has been unsubscribed'
-                                        ' from this list.') % email)
+            response = self.mailing_list.unsubscribe(email)
+            if response is not None and response.get('token') is not None:
+                messages.success(
+                    request, _('Your unsubscription request has been'
+                               ' submitted and is waiting for moderator'
+                               ' approval.'))
+            else:
+                messages.success(request, _('%s has been unsubscribed'
+                                            ' from this list.') % email)
         except ValueError as e:
             messages.error(request, e)
         return redirect('list_summary', self.mailing_list.list_id)
@@ -432,15 +595,31 @@ def list_mass_subscribe(request, list_id):
                     # Parse the data to get the address and the display name
                     display_name, address = email.utils.parseaddr(data)
                     validate_email(address)
+                    p_v = form.cleaned_data['pre_verified']
+                    p_c = form.cleaned_data['pre_confirmed']
+                    p_a = form.cleaned_data['pre_approved']
+                    inv = form.cleaned_data['invitation']
                     mailing_list.subscribe(
                         address=address,
                         display_name=display_name,
-                        pre_verified=form.cleaned_data['pre_verified'],
-                        pre_confirmed=form.cleaned_data['pre_confirmed'],
-                        pre_approved=form.cleaned_data['pre_approved'])
+                        pre_verified=p_v,
+                        pre_confirmed=p_c,
+                        pre_approved=p_a,
+                        invitation=inv,
+                        send_welcome_message=form.cleaned_data[
+                            'send_welcome_message'])
+                    if inv:
+                        message = _('The address %(address)s has been'
+                                    ' invited to %(list)s.')
+                    else:
+                        message = _('The address %(address)s has been'
+                                    ' subscribed to %(list)s.')
+                        if not (p_v and p_c and p_a):
+                            message += _(' Subscription may be pending address'
+                                         ' verification, confirmation or'
+                                         ' approval as appropriate.')
                     messages.success(
-                        request, _('The address %(address)s has been'
-                                   ' subscribed to %(list)s.') %
+                        request, message %
                         {'address': address,
                          'list': mailing_list.fqdn_listname})
                 except HTTPError as e:
@@ -448,6 +627,9 @@ def list_mass_subscribe(request, list_id):
                 except ValidationError:
                     messages.error(request, _('The email address %s'
                                               ' is not valid.') % address)
+                    # When the email address is bad, set the focus back to the
+                    # email field after returning to the same page.
+                    form.fields['emails'].widget.attrs['autofocus'] = True
     else:
         form = ListMassSubscription()
     return render(request, 'postorius/lists/mass_subscribe.html',
@@ -455,7 +637,6 @@ def list_mass_subscribe(request, list_id):
 
 
 class ListMassRemovalView(MailingListView):
-
     """Class For Mass Removal"""
 
     @method_decorator(login_required)
@@ -471,20 +652,31 @@ class ListMassRemovalView(MailingListView):
         if not form.is_valid():
             messages.error(request, _('Please fill out the form correctly.'))
         else:
-            for address in form.cleaned_data['emails']:
+            valid_emails = []
+            invalid_emails = []
+            for data in form.cleaned_data['emails']:
                 try:
+                    # Parse the data to get the address.
+                    address = email.utils.parseaddr(data)[1]
                     validate_email(address)
-                    self.mailing_list.unsubscribe(address.lower())
-                    messages.success(
-                        request, _('The address %(address)s has been'
-                                   ' unsubscribed from %(list)s.') %
-                        {'address': address,
-                         'list': self.mailing_list.fqdn_listname})
-                except (HTTPError, ValueError) as e:
-                    messages.error(request, e)
+                    valid_emails.append(address)
                 except ValidationError:
-                    messages.error(request, _('The email address %s'
-                                              ' is not valid.') % address)
+                    invalid_emails.append(data)
+
+            try:
+                self.mailing_list.mass_unsubscribe(valid_emails)
+                messages.success(
+                    request, _('These addresses {address} have been'
+                               ' unsubscribed from {list}.'.format(
+                                   address=', '.join(valid_emails),
+                                   list=self.mailing_list.fqdn_listname)))
+            except (HTTPError, ValueError) as e:
+                messages.error(request, e)
+
+            if len(invalid_emails) > 0:
+                messages.error(
+                    request,
+                    _('The email address %s is not valid.') % invalid_emails)
         return redirect('mass_removal', self.mailing_list.list_id)
 
 
@@ -562,18 +754,18 @@ def moderate_held_message(request, list_id):
 
     moderation_choices = dict(ACTION_CHOICES)
     if moderation_choice in moderation_choices:
-        try:
-            member = mailing_list.get_member(msg.sender)
+        member = get_member_or_nonmember(mailing_list, msg.sender)
+        if member is not None:
             member.moderation_action = moderation_choice
             member.save()
             messages.success(
                 request,
                 _('Moderation action for {} set to {}'.format(
                     member, moderation_choices[moderation_choice])))
-        except ValueError as e:
+        else:
             messages.error(
                 request,
-                _('Failed to set moderation action: {}'.format(e)))
+                _('Failed to set moderation action for {}'.format(msg.sender)))
     return redirect('list_held_messages', list_id)
 
 
@@ -718,24 +910,19 @@ def list_index_authenticated(request):
     client = get_mailman_client()
     choosable_domains = _get_choosable_domains(request)
 
-    # Get all the verified addresses of the user.
-    user_emails = EmailAddress.objects.filter(
-        user=request.user, verified=True).order_by(
-            "email").values_list("email", flat=True)
+    # Get the user_id of the current user
+    user_id = get_mailman_user_id(request.user)
 
-    # Get all the mailing lists for the current user.
-    all_lists = []
     mail_host = _get_mail_host(request.get_host().split(':')[0])
-    for user_email in user_emails:
-        try:
-            all_lists.extend(
-                client.find_lists(user_email,
-                                  role=role,
-                                  mail_host=mail_host,
-                                  count=sys.maxsize))
-        except HTTPError:
-            # No lists exist with the given role for the given user.
-            pass
+    # Get all the mailing lists for the current user.
+    try:
+        all_lists = client.find_lists(user_id,
+                                      role=role,
+                                      mail_host=mail_host,
+                                      count=sys.maxsize)
+    except HTTPError:
+        # No lists exist with the given role for the given user.
+        all_lists = []
     # If the user has no list that they are subscriber/owner/moderator of, we
     # just redirect them to the index page with all lists.
     if len(all_lists) == 0 and role is None:
@@ -778,13 +965,21 @@ def list_index(request, template='postorius/index.html'):
         _get_list_page, request.GET.get('page'), request.GET.get('count'),
         paginator_class=MailmanPaginator)
 
-    choosable_domains = _get_choosable_domains(request)
+    # This is just an optimization to skip un-necessary API
+    # calls. postorius/index.html page shows the 'Create New Domain' button
+    # when the logged-in user is a super user. There is no point making those
+    # API calls if the user isn't a Superuser. So, just call the number 0 if
+    # the user isn't SU.
+    if request.user.is_superuser:
+        domain_count = len(_get_choosable_domains(request))
+    else:
+        domain_count = 0
 
     return render(request, template,
                   {'lists': lists,
                    'check_advertised': True,
                    'all_lists': True,
-                   'domain_count': len(choosable_domains)})
+                   'domain_count': domain_count})
 
 
 @login_required
@@ -833,11 +1028,27 @@ def list_subscription_requests(request, list_id):
         )
 
 
-def _list_subscriptions(request, list_id, token_owner, template, page_title):
+@login_required
+@list_moderator_required
+def list_unsubscription_requests(request, list_id):
+    """Shows a list of pending unsubscription requests."""
+    return _list_subscriptions(
+        request=request,
+        list_id=list_id,
+        token_owner=TokenOwner.moderator,
+        template='postorius/lists/subscription_requests.html',
+        page_title=_('Un-subscriptions pending approval'),
+        request_type='unsubscription'
+        )
+
+
+def _list_subscriptions(
+        request, list_id, token_owner, template, page_title,
+        request_type='subscription'):
     m_list = List.objects.get_or_404(fqdn_listname=list_id)
     requests = [req
-                for req in m_list.requests
-                if req['token_owner'] == token_owner]
+                for req in m_list.get_requests(
+                    token_owner=token_owner, request_type=request_type)]
     paginated_requests = paginate(
         requests,
         request.GET.get('page'),
@@ -852,7 +1063,8 @@ def _list_subscriptions(request, list_id, token_owner, template, page_title):
 
 @login_required
 @list_moderator_required
-def handle_subscription_request(request, list_id, request_id, action):
+@require_http_methods(['POST'])
+def handle_subscription_request(request, list_id, request_id):
     """
     Handle a subscription request. Possible actions:
         - accept
@@ -866,11 +1078,19 @@ def handle_subscription_request(request, list_id, request_id, action):
         'discard': _('The request has been discarded.'),
         'defer': _('The request has been defered.'),
     }
-    assert action in confirmation_messages
+    # Get the action by the presence of one of the values in POST data.
+    for each in confirmation_messages:
+        if each in request.POST:
+            action = each
+    # If None of the actions were specified, just raise an error and return.
+    m_list = List.objects.get_or_404(fqdn_listname=list_id)
+    if action is None:
+        messages.error(f'Invalid or missing action {action!r}')
+        return redirect('list_subscription_requests', m_list.list_id)
+    reason = request.POST.get('reason')
     try:
-        m_list = List.objects.get_or_404(fqdn_listname=list_id)
         # Moderate request and add feedback message to session.
-        m_list.moderate_request(request_id, action)
+        m_list.moderate_request(request_id, action, reason)
         messages.success(request, confirmation_messages[action])
     except HTTPError as e:
         if e.code == 409:
@@ -940,7 +1160,20 @@ def list_settings(request, list_id=None, visible_section=None,
                     if key in form_class.mlist_properties:
                         setattr(m_list, key, form.cleaned_data[key])
                     else:
-                        list_settings[key] = form.cleaned_data[key]
+                        val = form.cleaned_data[key]
+                        # Empty list isn't really a valid value and Core
+                        # interprets empty string as empty value for
+                        # ListOfStringsField. We are doing it here instead of
+                        # ListOfStringsField.to_python() because output from
+                        # list of strings is expected to be a list and the NULL
+                        # value is hence an empty list. The serialization of
+                        # the empty list is for us empty string, hence we are
+                        # doing this here. Technically, it can be done outside
+                        # of this view, but there aren't any other use cases
+                        # where we'd want an empty list of strings.
+                        if val == []:
+                            val = ''
+                        list_settings[key] = val
                 list_settings.save()
                 messages.success(request, _('The settings have been updated.'))
                 mailinglist_modified.send(sender=List, list_id=m_list.list_id)
@@ -1006,7 +1239,6 @@ def remove_role(request, list_id=None, role=None, address=None,
 @login_required
 @list_owner_required
 def remove_all_subscribers(request, list_id):
-
     """Empty the list by unsubscribing all members."""
 
     mlist = List.objects.get_or_404(fqdn_listname=list_id)
@@ -1103,3 +1335,67 @@ def list_header_matches(request, list_id):
         'list': m_list,
         'formset': formset,
         })
+
+
+def confirm_token(request, list_id):
+    """Confirm token sent via Mailman Core.
+
+    This view allows confirming tokens sent via Mailman Core for confirming
+    subscriptions or verifying email addresses. This is an un-authenticated
+    view and the authentication is done by assuming the secrecy of the token
+    sent.
+
+    :param request: The Django request object.
+    :param list_id: The current mailinglist id.
+
+    """
+    m_list = List.objects.get_or_404(fqdn_listname=list_id)
+    if request.method == 'POST':
+        form = TokenConfirmForm(request.POST)
+        # If the token is None or something, just raise 400 error.
+        if not form.is_valid():
+            raise HTTPError(request.path, 400, _('Invalid confirmation token'),
+                            None, None)
+        token = form.cleaned_data.get('token')
+        try:
+            pending_req = m_list.get_request(token)
+        except HTTPError as e:
+            if e.code == 404:
+                raise HTTPError(request.path, 404,
+                                _('Token expired or invalid.'), None, None)
+            raise
+        # Since we only accept the token there isn't any need for Form data. We
+        # just need a POST request at this URL to accept the token.
+        m_list.moderate_request(token, action='accept')
+        return redirect('list_summary', m_list.list_id)
+    # Get the token from url parameter.
+    token = request.GET.get('token')
+    try:
+        pending_req = m_list.get_request(token)
+    except HTTPError as e:
+        if e.code == 404:
+            raise HTTPError(request.path, 404, _('Token expired or invalid.'),
+                            None, None)
+        raise
+
+    # If this is a token pending moderator approval, they need to login and
+    # approve it from the pending requests page.
+    if pending_req.get('token_owner') == 'moderator':
+        return redirect('list_subscription_requests', m_list.list_id)
+
+    token_type = pending_req.get('type')
+    form = TokenConfirmForm(initial=dict(token=token))
+    # Show the display_name if it is not None or "".
+    if pending_req.get('display_name'):
+        addr = email.utils.formataddr((pending_req.get('display_name'),
+                                       pending_req.get('email')))
+    else:
+        addr = pending_req.get('email')
+
+    return render(request, 'postorius/lists/confirm_token.html', {
+        'mlist': m_list,
+        'addr': addr,
+        'token': token,
+        'type': token_type,
+        'form': form,
+    })
